#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 29 13:59:06 2021

@author: jpark
"""


import numpy as np

from astropy.coordinates import EarthLocation
import astropy.time as at
import datetime as dt


from AIPS import AIPS
from AIPSTask import AIPSTask
from AIPSData import AIPSUVData, AIPSImage
from Wizardry.AIPSData import AIPSUVData as WAIPSUVData

import aipsutil as au

from os import path

from IPython import embed


def uvprt(data, select):
    """
    Extract UV data from ParselTongue UVData.
    
    Args:
        data (ParselTongue UVData): an input ParselTongue UVData.
        select (str): type of the data that will be extracted.
    
    Returns:
        list(s) of the selected UV data.
    """        
    
    ifnum = data.header.naxis[3]
    
    dumu, dumv, ifarr, time, ant1, ant2, rrreal, rrimag, rrweight, llreal, llimag, llweight, rlreal, rlimag, rlweight, lrreal, lrimag, lrweight = \
        [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], [], []
    
    for ifn in range(ifnum):
        for visibility in data:
            if((visibility.visibility[ifn,0,0,2] > 0.) & (visibility.visibility[ifn,0,1,2] > 0.) & (visibility.visibility[ifn,0,2,2] > 0.) & (visibility.visibility[ifn,0,3,2] > 0.)):
                dumu.append(visibility.uvw[0])
                dumv.append(visibility.uvw[1])
                ifarr.append(ifn+1)
                time.append(visibility.time)
                rrreal.append(visibility.visibility[ifn,0,0,0])
                rrimag.append(visibility.visibility[ifn,0,0,1])
                rrweight.append(visibility.visibility[ifn,0,0,2])
                llreal.append(visibility.visibility[ifn,0,1,0])
                llimag.append(visibility.visibility[ifn,0,1,1])
                llweight.append(visibility.visibility[ifn,0,1,2])
                rlreal.append(visibility.visibility[ifn,0,2,0])
                rlimag.append(visibility.visibility[ifn,0,2,1])
                rlweight.append(visibility.visibility[ifn,0,2,2])
                lrreal.append(visibility.visibility[ifn,0,3,0])
                lrimag.append(visibility.visibility[ifn,0,3,1])
                lrweight.append(visibility.visibility[ifn,0,3,2])
                ant1.append(visibility.baseline[0])
                ant2.append(visibility.baseline[1])
        
    if(np.sum(rrreal) == 0.):
        return
    
    selectarr = ["u", "v", "ifarr", "time", "ant1", "ant2", "rrreal", "rrimag", "rrweight", "llreal", "llimag", "llweight", "rlreal", "rlimag", "rlweight", "lrreal", "lrimag", "lrweight"]
    package = [dumu, dumv, ifarr, time, ant1, ant2, rrreal, rrimag, rrweight, llreal, llimag, llweight, rlreal, rlimag, rlweight, lrreal, lrimag, lrweight]
    
    for i in range(len(selectarr)):
        if(select == selectarr[i]): return package[i]
    
    if(select == "all"): return dumu, dumv, ifarr, time, ant1, ant2, rrreal, rrimag, rrweight, llreal, llimag, llweight, rlreal, rlimag, rlweight, lrreal, lrimag, lrweight
    


def pol_model_uvprt(data, ifn, select):
    """
    Extract UV data from ParselTongue UVData for instrumental polarization self-calibration.
    
    Args:
        data (ParselTongue UVData): an input ParselTongue UVData.
        ifn (int): selected IF number
        select (str): type of the data that will be extracted.
    
    Returns:
        list(s) of the selected UV data.
    """        
    real, imag = [], []

#        (visibility.visibility[ifn,0,0,2] == np.nan) | (visibility.visibility[ifn,0,1,2] == np.nan) | (visibility.visibility[ifn,0,2,2] == np.nan) | (visibility.visibility[ifn,0,3,2] == np.nan))
    for visibility in data:
        if((visibility.visibility[ifn,0,0,2] > 0.) & (visibility.visibility[ifn,0,1,2] > 0.) & (visibility.visibility[ifn,0,2,2] > 0.) & (visibility.visibility[ifn,0,3,2] > 0.)):
            real.append(visibility.visibility[ifn,0,0,0])
            imag.append(visibility.visibility[ifn,0,0,1])
    
    if(np.sum(real) == 0.):
        return None, None
            
    selectarr = ["real", "imag"]
    package = [real, imag]
    
    for i in range(len(selectarr)):
        if(select == selectarr[i]): return package[i]
    
    if(select == "all"): return real, imag
    


#    def get_parang(self, time, ant, sourcearr, source, obsra, obsdec):
def get_parang(yeararr, montharr, dayarr, time, raarr, decarr, lonarr, latarr, f_el_arr, f_par_arr, phi_off_arr): # Version 1.1!
    """
    Calculate antenna field-rotation angles.
    
    Args:
        time (numpy.array): a numpy array of time in UTC of the visibilities.
        ant (numpy.array): a numpy array of antenna number of the visibilities.
        sourcearr (numpy.array): a numpy array of source of the visibilities.
        source (list): a list of calibrators.
        obsra (list): a list of calibrators' right ascension in units of degrees.
        obsdec (list): a list of calibrators' declination in units of degrees.
    
    Returns:
        a numpy of the field-rotation angles.
    """        
    
    
    latarr, decarr = np.radians(latarr), np.radians(decarr)

    hour = np.floor(time)
    minute = np.floor((time - hour) * 60.)
    second = (time - hour - minute / 60.) * 3600.
    
    hour = hour.astype('int')
    minute = minute.astype('int')
    second = second.astype('int')
    
    dumdatetime = [dt.datetime(a, b, c, d, e, f) for a, b, c, d, e, f in zip(yeararr, montharr, dayarr, hour, minute, second)]

    dumt = [] 

    for dum in dumdatetime:
        dumt.append("{:04d}-{:02d}-{:02d}T{:02d}:{:02d}:{:f}".format(dum.year, dum.month, dum.day, dum.hour, dum.minute, dum.second + dum.microsecond * 1e-6))

    dumt = at.Time(dumt)
    gst = dumt.sidereal_time('mean','greenwich').hour

    # Obtain field-rotation angles using the known equations.
    hangle = np.radians(gst * 15. + lonarr - raarr)        
    parang = np.arctan2((np.sin(hangle) * np.cos(latarr)), (np.sin(latarr) * np.cos(decarr) - np.cos(latarr) * np.sin(decarr) * np.cos(hangle)))        
    altitude = np.arcsin(np.sin(decarr) * np.sin(latarr) + np.cos(decarr) * np.cos(latarr) * np.cos(hangle))
    pang = f_el_arr * altitude + f_par_arr * parang + phi_off_arr
    
    
    return pang
    

# #    def get_parang(self, time, ant, sourcearr, source, obsra, obsdec):
# def get_parang(time, ant, sourcearr, year, month, day, source, obsra, obsdec, longarr, latiarr, f_el, f_par, phi_off): # Version 1.1!
#     """
#     Calculate antenna field-rotation angles.
    
#     Args:
#         time (numpy.array): a numpy array of time in UTC of the visibilities.
#         ant (numpy.array): a numpy array of antenna number of the visibilities.
#         sourcearr (numpy.array): a numpy array of source of the visibilities.
#         source (list): a list of calibrators.
#         obsra (list): a list of calibrators' right ascension in units of degrees.
#         obsdec (list): a list of calibrators' declination in units of degrees.
    
#     Returns:
#         a numpy of the field-rotation angles.
#     """        
    
#     nant = np.max(ant) + 1
#     num = len(time)
    
#     lonarr, latarr, raarr, decarr, elarr, pararr, phiarr, yeararr, montharr, dayarr = \
#         np.zeros(num), np.zeros(num), np.zeros(num), np.zeros(num), np.zeros(num), np.zeros(num), np.zeros(num), np.zeros(num), np.zeros(num), np.zeros(num) # Version 1.1!
        
#     # Produce numpy arrays for antenna longitudes, latitudes, and the coefficients of the field-rotation angle equations.
#     for m in range(nant):
#         lonarr[(ant == m)] = longarr[m]
#         latarr[(ant == m)] = latiarr[m]
#         elarr[(ant == m)] = f_el[m]
#         pararr[(ant == m)] = f_par[m]
#         phiarr[(ant == m)] = phi_off[m]
    
#     # Produce numpy arrays for sources RA and Dec.
#     for l in range(len(source)):
#         raarr[sourcearr == source[l]] = obsra[l]
#         decarr[sourcearr == source[l]] = obsdec[l]
#         yeararr[sourcearr == source[l]] = year[l]
#         montharr[sourcearr == source[l]] = month[l]
#         dayarr[sourcearr == source[l]] = day[l]
        
#     latarr, decarr = np.radians(latarr), np.radians(decarr)

    
#     hour = np.floor(time)
#     minute = np.floor((time - hour) * 60.)
#     second = (time - hour - minute / 60.) * 3600.
    
#     for i in range(100):
#         dayarr[hour>=24.] += 1. # Version 1.1!
#         hour[hour>=24.] -= 24. # Version 1.1!
        
#     # Convert UTC to GST using astropy Time.
# #        dumt = at.Time(["{:04d}-{:02d}-{:02d}T{:02d}:{:02d}:{:f}".format(self.year, self.month, self.day+int(dt), int(hr), int(mn), sec) for dt, hr, mn, sec in zip(date, hour, minute, second)])
    
#     dumt = at.Time(["{:04d}-{:02d}-{:02d}T{:02d}:{:02d}:{:f}".format(int(yr), int(mo), int(dt), int(hr), int(mn), sec) for yr, mo, dt, hr, mn, sec in zip(yeararr, montharr, dayarr, hour, minute, second)]) # Version 1.1!
#     gst = dumt.sidereal_time('mean','greenwich').hour
    
    
#     # Obtain field-rotation angles using the known equations.
#     hangle = np.radians(gst * 15. + lonarr - raarr)        
#     parang = np.arctan2((np.sin(hangle) * np.cos(latarr)), (np.sin(latarr) * np.cos(decarr) - np.cos(latarr) * np.sin(decarr) * np.cos(hangle)))        
#     altitude = np.arcsin(np.sin(decarr) * np.sin(latarr) + np.cos(decarr) * np.cos(latarr) * np.cos(hangle))
#     pang = elarr * altitude + pararr * parang + phiarr
        
    
#     return pang
    


def coord(antname, antx, anty, antz):
    """
    Convert antenna positions from Cartesian to spherical coordinates using astropy.
    
    Returns:
        lists of antenna longitudes, latitudes, and heights.
    """      
    lonarr = []
    latarr = []
    heightarr = []
    
    for i in range(len(antname)):
        lonarr.append(EarthLocation.from_geocentric(antx[i], anty[i], antz[i], unit = 'm').to_geodetic()[0].value)
        latarr.append(EarthLocation.from_geocentric(antx[i], anty[i], antz[i], unit = 'm').to_geodetic()[1].value)
        heightarr.append(EarthLocation.from_geocentric(antx[i], anty[i], antz[i], unit = 'm').to_geodetic()[2].value)
        
    
    return lonarr, latarr, heightarr


def calendar(sourcearr, calsour, year, month, day, obsra, obsdec):
    dnum = len(sourcearr)
    yeararr, montharr, dayarr, raarr, decarr = np.zeros(dnum, dtype = 'int'), np.zeros(dnum, dtype = 'int'), np.zeros(dnum, dtype = 'int'), np.zeros(dnum, dtype = 'float'), np.zeros(dnum, dtype = 'float')
    for l in range(len(calsour)):
        yeararr[sourcearr == calsour[l]] = year[l]
        montharr[sourcearr == calsour[l]] = month[l]
        dayarr[sourcearr == calsour[l]] = day[l]
        raarr[sourcearr == calsour[l]] = obsra[l]
        decarr[sourcearr == calsour[l]] = obsdec[l]
    
    return yeararr, montharr, dayarr, raarr, decarr


def coordarr(longi, lati, f_el, f_par, phi_off, antarr):
    longi = np.array(longi)
    lati = np.array(lati)
    f_el = np.array(f_el)
    f_par = np.array(f_par)
    phi_off = np.array(phi_off)
            
    longarr = longi[antarr]
    latarr = lati[antarr]
    f_el = f_el[antarr]
    f_par = f_par[antarr]
    phi_off = phi_off[antarr]
    
    return longarr, latarr, f_el, f_par, phi_off


def basic_info(source, direc, dataname):
    obsra, obsdec, year, month, day = [], [], [], [], []
    
    for l in range(len(source)):
        inname = str(source[l])
        
        data = AIPSUVData(inname, 'EDIT', 1, 1)
        if(data.exists() == True):
            data.clrstat()
            data.zap()
        
        # Load UVFITS files.
        au.runfitld(inname, 'EDIT', direc + dataname + source[l] + '.uvf')
        
        data = AIPSUVData(inname, 'EDIT', 1, 1)
        
        dum_obsra, dum_obsdec = get_obscoord(data)
        
        # Extract source coordinates from the header.
        obsra.append(dum_obsra)
        obsdec.append(dum_obsdec)

        
        dumyear, dummonth, dumday = get_obsdate(data)
        
        year.append(dumyear)
        month.append(dummonth)
        day.append(dumday)
        
        # Extract antenna, frequency, mount information, etc, from the header.
        if(l == 0):
            
            antname, antx, anty, antz, antmount, f_par, f_el, phi_off = get_antcoord(data)
            
            ifnum, freq = get_freqinfo(data)
        
        data.zap()
        
    info = {"obsra": obsra, "obsdec": obsdec, "year": year, "month": month, "day": day, "antname": antname, "antx": antx, "anty": anty, "antz": antz, "antmount": antmount, "ifnum": ifnum, "freq": freq, \
            "f_par": f_par, "f_el": f_el, "phi_off": phi_off}
    
    return info



def get_obsdate(data):
    
    obsdate = data.header.date_obs
    year = int(obsdate[0:4])
    month = int(obsdate[5:7])
    day = int(obsdate[8:10])
    
    return year, month, day


def get_obscoord(data):
    
    obsra = data.header.crval[4]
    obsdec = data.header.crval[5]
    
    return obsra, obsdec


def get_antcoord(data):
    
    antname = []
    antx = []
    anty = []
    antz = []
    antmount = []
    
    antable = data.table('AN', 1)
    for row in antable:
        antname.append(row.anname.replace(' ', ''))
        antx.append(row.stabxyz[0])
        anty.append(row.stabxyz[1])
        antz.append(row.stabxyz[2])
        antmount.append(row.mntsta)

    f_par = []
    f_el = []
    phi_off = []
    
    for st in range(len(antname)):
        f_par.append(1.)
        if(antmount[st] == 0) | (antmount[st] == 4) | (antmount[st] == 5):
            f_el.append(0.)
        if(antmount[st] == 4):
            f_el.append(1.)
        if(antmount[st] == 5):
            f_el.append(-1.)
        phi_off.append(0.)
        
    return antname, antx, anty, antz, antmount, f_par, f_el, phi_off


def get_freqinfo(data):
        
    fqtable = data.table('FQ', 1)
    if(isinstance(fqtable[0].if_freq, float) == True):
        ifnum = 1
        IFfreq = [data.header.crval[2] / 1e9]
        freq = "{0:.3f}".format(IFfreq[0]) + ' GHz'
    else:
        ifnum = len(fqtable[0].if_freq)
        freq = [str((it + data.header.crval[2]) / 1e9) + ' GHz' for it in fqtable[0].if_freq]
    
    return ifnum, freq


def pd_modifier(data):
    
    # Make new columns for amplitudes, phases, and corresponding errors, etc.
    data.loc[:,"rrsigma"], data.loc[:,"llsigma"], data.loc[:,"rlsigma"], data.loc[:,"lrsigma"] = \
        1. / data.loc[:,"rrweight"] ** (0.5), 1. / data.loc[:,"llweight"], 1. / data.loc[:,"rlweight"] ** (0.5), 1. / data.loc[:,"lrweight"] ** (0.5)
    
    data.loc[:,"rramp"], data.loc[:,"llamp"], data.loc[:,"rlamp"], data.loc[:,"lramp"] = \
        np.absolute(data.loc[:,"rrreal"] + 1j*data.loc[:,"rrimag"]), np.absolute(data.loc[:,"llreal"] + 1j*data.loc[:,"llimag"]), \
        np.absolute(data.loc[:,"rlreal"] + 1j*data.loc[:,"rlimag"]), np.absolute(data.loc[:,"lrreal"] + 1j*data.loc[:,"lrimag"])
    
    data.loc[:,"rrphas"], data.loc[:,"llphas"], data.loc[:,"rlphas"], data.loc[:,"lrphas"] = \
        np.angle(data.loc[:,"rrreal"] + 1j*data.loc[:,"rrimag"]), np.angle(data.loc[:,"llreal"] + 1j*data.loc[:,"llimag"]), \
        np.angle(data.loc[:,"rlreal"] + 1j*data.loc[:,"rlimag"]), np.angle(data.loc[:,"lrreal"] + 1j*data.loc[:,"lrimag"])
    
    data.loc[:,"rramp_sigma"], data.loc[:,"llamp_sigma"], data.loc[:,"rlamp_sigma"], data.loc[:,"lramp_sigma"] = \
        data.loc[:,"rrsigma"], data.loc[:,"llsigma"], data.loc[:,"rlsigma"], data.loc[:,"lrsigma"]
    
    data.loc[:,"rrphas_sigma"], data.loc[:,"llphas_sigma"], data.loc[:,"rlphas_sigma"], data.loc[:,"lrphas_sigma"] = \
        data.loc[:,"rrsigma"] / np.abs(data.loc[:,"rrreal"] + 1j*data.loc[:,"rrimag"]), \
        data.loc[:,"llsigma"] / np.abs(data.loc[:,"llreal"] + 1j*data.loc[:,"llimag"]), \
        data.loc[:,"rlsigma"] / np.abs(data.loc[:,"rlreal"] + 1j*data.loc[:,"rlimag"]), \
        data.loc[:,"lrsigma"] / np.abs(data.loc[:,"lrreal"] + 1j*data.loc[:,"lrimag"])
        
        # phaserror(data.loc[:,"rrreal"] + 1j*data.loc[:,"rrimag"], data.loc[:,"rrsigma"] + 1j*data.loc[:,"rrsigma"]), \
        # phaserror(data.loc[:,"llreal"] + 1j*data.loc[:,"llimag"], data.loc[:,"llsigma"] + 1j*data.loc[:,"llsigma"]), \
        # phaserror(data.loc[:,"rlreal"] + 1j*data.loc[:,"rlimag"], data.loc[:,"rlsigma"] + 1j*data.loc[:,"rlsigma"]), \
        # phaserror(data.loc[:,"lrreal"] + 1j*data.loc[:,"lrimag"], data.loc[:,"lrsigma"] + 1j*data.loc[:,"lrsigma"])
        
    
    dumrl, dumlr = data.loc[:,"rlreal"] + 1j*data.loc[:,"rlimag"], data.loc[:,"lrreal"] + 1j*data.loc[:,"lrimag"]
    dumrlsigma, dumlrsigma = data.loc[:,"rlsigma"] + 1j*data.loc[:,"rlsigma"], data.loc[:,"lrsigma"] + 1j*data.loc[:,"lrsigma"]
    
    dumq, dumu = (dumrl + dumlr) / 2., -1j * (dumrl - dumlr) / 2.
    dumqsigma, dumusigma = np.sqrt(dumrlsigma**2 + dumlrsigma**2) / 2., np.sqrt(dumrlsigma**2 + dumlrsigma**2) / 2.
    
    data.loc[:,"qamp"], data.loc[:,"uamp"], data.loc[:,"qphas"], data.loc[:,"uphas"] = np.absolute(dumq), np.absolute(dumu), np.angle(dumq), np.angle(dumu)
    data.loc[:,"qphas_sigma"], data.loc[:,"uphas_sigma"] = np.real(dumqsigma) / np.abs(dumq), np.real(dumusigma) / np.abs(dumu)
    data.loc[:,"qsigma"], data.loc[:,"usigma"] = np.real(dumqsigma), np.real(dumusigma)

    data["IF"] = data["IF"].astype('int32')
    data["ant1"] = data["ant1"].astype('int32')
    data["ant2"] = data["ant2"].astype('int32')
    data["pang1"] = data["pang1"].astype('float64')
    data["pang2"] = data["pang2"].astype('float64')
    
    return data


def get_model(data, direc, dataname, calsour, polcal_unpol, ifnum, pol_IF_combine, outputname = None, selfcal = False):
    """
    Extract Stokes Q and U visibility models and append them to the pandas dataframe.
    
    """
    
    # self.logger.info('\nGetting source polarization models for {:d} sources for {:d} IFs...'.format(len(self.polcalsour), self.ifnum))
        
    # AIPS.userno = self.aips_userno
    
    # if self.aipslog:
    #     AIPS.log = open(self.logfile, 'a')
    # AIPSTask.msgkill = -1
            
    
    mod_qrealarr, mod_qimagarr, mod_urealarr, mod_uimagarr = [], [], [], []
    
    
    for l in range(len(calsour)):
    
        inname = str(calsour[l])
    
        calib = AIPSUVData(inname, 'EDIT', 1, 1)
        if(calib.exists() == True):
            calib.clrstat()
            calib.zap()
        
        if selfcal:
            au.runfitld(inname, 'EDIT', direc + dataname + calsour[l]+'.calib')
        else:
            au.runfitld(inname, 'EDIT', direc + dataname + calsour[l]+'.uvf')
    
        calib = AIPSUVData(inname, 'EDIT', 1, 1)
        
        if polcal_unpol[l]:
            dumdata = data
            dumsource = np.array(dumdata.loc[:,"source"])
            dumlen = np.sum(dumsource == calsour[l])
            mod_qrealarr = mod_qrealarr + [0.] * dumlen
            mod_qimagarr = mod_qimagarr + [0.] * dumlen
            mod_urealarr = mod_urealarr + [0.] * dumlen
            mod_uimagarr = mod_uimagarr + [0.] * dumlen
       
        else:
            for k in range(ifnum):
                if(np.sum(data.loc[:, "IF"] == k+1) == 0):
                    continue

                qmap = AIPSImage(inname, 'QMAP', 1, 1)
                if(qmap.exists() == True):
                    qmap.clrstat()
                    qmap.zap()

                if pol_IF_combine:
                    fitsname = direc+dataname+calsour[l]+'.allIF.q.fits'
                else:
                    fitsname = direc+dataname+calsour[l]+'.IF'+str(k+1)+'.q.fits'
                
                
                # if not path.exists(fitsname):
                #     dum = 0
                #     calib = WAIPSUVData(inname, 'EDIT', 1, 1)
                #     for visibility in calib:
                #         if((visibility.visibility[k,0,0,2] > 0) & (visibility.visibility[k,0,1,2] > 0) & (visibility.visibility[k,0,2,2] > 0) & (visibility.visibility[k,0,3,2] > 0)):
                #             dum += 1 
    
                #     mod_qrealarr = mod_qrealarr + [0.] * dum
                #     mod_qimagarr = mod_qimagarr + [0.] * dum
                    
                #     calib = AIPSUVData(inname, 'EDIT', 1, 1)
                if not path.exists(fitsname):
                    raise Exception("The requested {:} file does not exist!".format(fitsname))
                    
                else:
                    
                    au.runfitld(inname, 'QMAP', fitsname)
                    qmap = AIPSImage(inname, 'QMAP', 1, 1)     
         
                    uvsub = AIPSUVData(inname, 'UVSUB', 1, 1)
                    if(uvsub.exists() == True):
                        uvsub.clrstat()
                        uvsub.zap()

                    au.runuvsub(inname, 'EDIT', 'QMAP', 1, 1)
                    
                    moddata = WAIPSUVData(inname, 'UVSUB', 1, 1)
                    mod_qreal, mod_qimag = pol_model_uvprt(moddata, k, "all")

                    if(mod_qreal != None):
                        mod_qrealarr = mod_qrealarr + mod_qreal
                        mod_qimagarr = mod_qimagarr + mod_qimag

                    moddata.zap()
                    qmap.zap()
                
                
                
                umap = AIPSImage(inname, 'UMAP', 1, 1)
                if(umap.exists() == True):
                    umap.clrstat()
                    umap.zap()


                if pol_IF_combine:
                    fitsname = direc+dataname+calsour[l]+'.allIF.u.fits'
                else:
                    fitsname = direc+dataname+calsour[l]+'.IF'+str(k+1)+'.u.fits'
                
                
                if not path.exists(fitsname):
                    dum = 0
                    calib = WAIPSUVData(inname, 'EDIT', 1, 1)
                    for visibility in calib:
                        if((visibility.visibility[k,0,0,2] > 0) & (visibility.visibility[k,0,1,2] > 0) & (visibility.visibility[k,0,2,2] > 0) & (visibility.visibility[k,0,3,2] > 0)):
                            dum += 1 
    
                    mod_urealarr = mod_urealarr + [0.] * dum
                    mod_uimagarr = mod_uimagarr + [0.] * dum
                    
                    calib = AIPSUVData(inname, 'EDIT', 1, 1)
                
                else:
                    au.runfitld(inname, 'UMAP', fitsname)
                    umap = AIPSImage(inname, 'UMAP', 1, 1)     
         
                    uvsub = AIPSUVData(inname, 'UVSUB', 1, 1)
                    if(uvsub.exists() == True):
                        uvsub.clrstat()
                        uvsub.zap()

                    au.runuvsub(inname, 'EDIT', 'UMAP', 1, 1)
                    
                    moddata = WAIPSUVData(inname, 'UVSUB', 1, 1)
                    mod_ureal, mod_uimag = pol_model_uvprt(moddata, k, "all")

                    if(mod_ureal != None):
                        mod_urealarr = mod_urealarr + mod_ureal
                        mod_uimagarr = mod_uimagarr + mod_uimag

                    moddata.zap()
                    umap.zap()
                    

        calib.zap()

    
    mod_qreal, mod_qimag, mod_ureal, mod_uimag = np.array(mod_qrealarr), np.array(mod_qimagarr), np.array(mod_urealarr), np.array(mod_uimagarr)
    
    
    mod_q, mod_u = mod_qreal + 1j*mod_qimag, mod_ureal + 1j*mod_uimag
    mod_rlreal, mod_rlimag, mod_lrreal, mod_lrimag = np.real(mod_q + 1j*mod_u), np.imag(mod_q + 1j*mod_u), np.real(mod_q - 1j*mod_u), np.imag(mod_q - 1j*mod_u)
    mod_rlamp, mod_rlphas, mod_lramp, mod_lrphas = np.absolute(mod_rlreal + 1j*mod_rlimag), np.angle(mod_rlreal + 1j*mod_rlimag), np.absolute(mod_lrreal + 1j*mod_lrimag), np.angle(mod_lrreal + 1j*mod_lrimag)
    mod_qamp, mod_qphas, mod_uamp, mod_uphas = np.absolute(mod_q), np.angle(mod_q), np.absolute(mod_u), np.angle(mod_u)
    

    # Append the model visibilities to the existing pandas dataframe as new columns.
    data.loc[:,"model_rlreal"], data.loc[:,"model_rlimag"], data.loc[:,"model_lrreal"], data.loc[:,"model_lrimag"], \
    data.loc[:,"model_rlamp"], data.loc[:,"model_rlphas"], data.loc[:,"model_lramp"], data.loc[:,"model_lrphas"], \
    data.loc[:,"model_qamp"], data.loc[:,"model_qphas"], data.loc[:,"model_uamp"], data.loc[:,"model_uphas"] = \
    mod_rlreal, mod_rlimag, mod_lrreal, mod_lrimag, mod_rlamp, mod_rlphas, mod_lramp, mod_lrphas, mod_qamp, mod_qphas, mod_uamp, mod_uphas
    
    return data


def evpacal(datain, dataout, clcorprm, logger = None): # Version 1.1
    
    if (logger != None):
        logger.info('Correcting EVPAs... \n Input file: {:} \n Output file: {:}'.format(datain, dataout))
    
    pinal = AIPSUVData('EVPA', 'PINAL', 1, 1)
    if(pinal.exists() == True):
        pinal.zap()
        
    au.runfitld('EVPA', 'PINAL', datain)
    
    pinal = AIPSUVData('EVPA', 'PINAL', 1, 1)
    
    aipssource = pinal.header.object
    
    multi = AIPSUVData('EVPA', 'MULTI', 1, 1)
    if(multi.exists() == True):
        multi.zap()
        
    au.runmulti('EVPA', 'PINAL')
    au.runclcor('EVPA', 'MULTI', 1, 1, clcorprm)
    
    pang = AIPSUVData(aipssource, 'PANG', 1, 1)
    if(pang.exists() == True):
        pang.zap()
        
    au.runsplitpang('EVPA')
            
    au.runfittp(aipssource, 'PANG', dataout)
    
    pinal.zap()
    multi.zap()
    pang.zap()



def get_scans(time, sourcearr, tsep = 2. / 60.):
    argsort = np.argsort(time)
    dum_sourcearr = sourcearr[argsort]
    dumx = np.sort(time)
    
    boundary_left = [np.min(dumx)]
    boundary_right = []
    boundary_source = [dum_sourcearr[0]]
    
    for j in range(len(dumx)-1):
        if(dumx[j+1] - dumx[j]) > tsep:
            boundary_left.append(dumx[j+1])
            boundary_right.append(dumx[j])
            boundary_source.append(dum_sourcearr[j+1])
            continue
        if(dum_sourcearr[j+1] != dum_sourcearr[j]):
            boundary_left.append(dumx[j+1])
            boundary_right.append(dumx[j])
            boundary_source.append(dum_sourcearr[j+1])
    
    boundary_right.append(np.max(dumx))
    
    return boundary_left, boundary_right, boundary_source


def bin_data(time, data, boundary_left, boundary_right, error = None, avg_nat = False):
    
    binx, biny, binsigma = [], [], []
    for i in range(len(boundary_left)):
        select = (time >= boundary_left[i]) & (time < boundary_right[i])
        if(np.sum(select) == 0): continue
        binx.append((boundary_left[i] + boundary_right[i]) / 2.)
        if (error is not None) & (avg_nat == True):
            biny.append(np.average(data[select], weights = 1. / error[select] ** 2, returned = True)[0])
            binsigma.append(np.abs(1. / np.average(data[select], weights = 1. / error[select] ** 2, returned = True)[1] ** 0.5))
        else:
            biny.append(np.mean(data[select]))
            binsigma.append(np.std(data[select]) / np.sqrt(float(np.sum(select))))
    
    binx, biny, binsigma = np.array(binx), np.array(biny), np.array(binsigma)
        
    return binx, biny, binsigma
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 29 13:59:06 2021

@author: jpark
"""


import numpy as np
import pandas as pd

from scipy.optimize import curve_fit

from astropy.coordinates import EarthLocation
import astropy.time as at

from AIPS import AIPS
from AIPSTask import AIPSTask
from AIPSData import AIPSUVData, AIPSImage
from Wizardry.AIPSData import AIPSUVData as WAIPSUVData

import aipsutil as au
import obshelpers as oh
import plothelpers as ph

import os
from os import path

from IPython import embed


from multiprocessing import cpu_count, Pool


max_count = cpu_count()
nproc = max_count - 1


def cleanqu(direc, data, mask, save, bif, eif, ms, ps, uvbin, uvpower, dynam, shift_x, shift_y, stokes, log = None): # Version 1.1
    """
    Perform imaging of Stokes Q and U in Difmap.
    
    Args:
        data (str): the name of the UVFITS file to be CLEANed.
        mask (str): the name of the CLEAN windows to be used.
        save (str): the name of the output Difmap save file.
        log (str): the name of the Difmap log file.
    """
    
    # logname = ''
    
    
    # Write a simple Difmap script for CLEAN in the working directory.
#        if not path.exists(self.direc+'GPCAL_Difmap_v1'):
    f = open(direc+'GPCAL_Difmap_v1','w')
    
    f.write('observe %1\nmapcolor rainbow, 1, 0.5\nselect %13, %2, %3\nmapsize %4, %5\nuvweight %6, %7\nrwin %8\nshift %9,%10\ndo i=1,100\nclean 100, 0.02, imstat(rms)*%11\nend do\nselect i\nsave %12.%13\nexit')
        
    f.close()
        
    
    curdirec = os.getcwd()
    
    logname = 'gpcal_difmap.log'

    os.chdir(direc)
    if log != None:
        command = "echo @GPCAL_Difmap_v1 %s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s | difmap > %s" %(data,bif,eif,ms,ps,uvbin,uvpower,mask,shift_x,shift_y,dynam,save,stokes,logname) # Version 1.1!
    else:
        command = "echo @GPCAL_Difmap_v1 %s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s | difmap > /dev/null 2>&1" %(data,bif,eif,ms,ps,uvbin,uvpower,mask,shift_x,shift_y,dynam,save,stokes) # Version 1.1!
    
    # log.info('\nMaking CLEAN models for Stokes Q & U maps for {:s}...'.format(data))
    
    os.system(command)
                
    if log != None:
        os.system('cat ' + logname + ' > ' + log)
    
    os.chdir(curdirec)
    


def cleanqu2(args): # Version 1.1
    cleanqu3(*args)

def cleanqu3(direc, data, mask, save, bif, eif, ms, ps, uvbin, uvpower, dynam, shift_x, shift_y, stokes, log): # Version 1.1
    cleanqu(direc, data, mask, save, bif, eif, ms, ps, uvbin, uvpower, dynam, shift_x, shift_y, stokes, log = None)
    

def cleanqu_run(args): # Version 1.1
    """
    Perform imaging of Stokes Q and U in Difmap.
    
    Args:
        data (str): the name of the UVFITS file to be CLEANed.
        mask (str): the name of the CLEAN windows to be used.
        save (str): the name of the output Difmap save file.
        log (str): the name of the Difmap log file.
    """

    pool = Pool(processes = nproc)
    pool.map(cleanqu2, args)
    pool.close()
    pool.join()
    
    # cleanqu(direc, data, mask, save, bif, eif, ms, ps, uvbin, uvpower, dynam, shift_x, shift_y, stokes, log = log)
    


def deq(x, *p):
    """
    The D-term models for the initial D-term estimation using the similarity assumption.
    
    Args:
        x: dummy parameters (not to be used).
        *p (args): the best-fit parameters args.
    """
    
    nant, cnum, calsour, modamp, modphas, pang1, pang2, ant1, ant2, sourcearr, rramp, rrphas, llamp, llphas, model_rlreal, model_rlimag, model_rlamp, model_rlphas, model_lrreal, model_lrimag, model_lramp, model_lrphas = x
    
    RiLj_Real = np.zeros(len(pang1))
    RiLj_Imag = np.zeros(len(pang1))
    LiRj_Real = np.zeros(len(pang1))
    LiRj_Imag = np.zeros(len(pang1))
    
    dump = np.array(p)
    
    dumreal = dump[2*ant1]
    dumimag = dump[2*ant1 + 1]
    Tot_D_iR_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_iR_phas = np.angle(dumreal + 1j*dumimag)

    dumreal = dump[2*nant + 2*ant2]
    dumimag = dump[2*nant + 2*ant2 + 1]
    Tot_D_jL_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_jL_phas = np.angle(dumreal + 1j*dumimag)
    
    dumreal = dump[2*nant + 2*ant1]
    dumimag = dump[2*nant + 2*ant1 + 1]
    Tot_D_iL_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_iL_phas = np.angle(dumreal + 1j*dumimag)
    
    dumreal = dump[2*ant2]
    dumimag = dump[2*ant2 + 1]
    Tot_D_jR_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_jR_phas = np.angle(dumreal + 1j*dumimag)


    P_ij_amp = [np.absolute(it1 + 1j*it2) for it1, it2 in zip([p[nant * 4 + s * 2] for s in range(sum(cnum))], [p[nant * 4 + s * 2 + 1] for s in range(sum(cnum))])]
    P_ij_phas = [np.angle(it1 + 1j*it2) for it1, it2 in zip([p[nant * 4 + s * 2] for s in range(sum(cnum))], [p[nant * 4 + s * 2 + 1] for s in range(sum(cnum))])]
    
    for l in range(len(calsour)):
        select = (sourcearr == calsour[l])
        
        if(cnum[l] != 0.):
            for t in range(cnum[l]):
                if(l==0):
                    dummodamp = np.array(modamp[t])
                    dummodphas = np.array(modphas[t])
                    Pick = t
                else:
                    dummodamp = np.array(modamp[sum(cnum[0:l])+t])
                    dummodphas = np.array(modphas[sum(cnum[0:l])+t])
                    Pick = sum(cnum[0:l]) + t
                    
                submodamp = dummodamp[select]
                submodphas = dummodphas[select]
                
        
                Pamp = P_ij_amp[Pick]
                Pphas = P_ij_phas[Pick]
                
    
                RiLj_Real[select] += Pamp * submodamp * np.cos(submodphas + Pphas) + \
                  Tot_D_iR_amp[select] * Tot_D_jL_amp[select] * Pamp * submodamp * np.cos(submodphas + Tot_D_iR_phas[select] - Tot_D_jL_phas[select] - Pphas + 2. * (pang1[select] + pang2[select]))
        
                RiLj_Imag[select] += Pamp * submodamp * np.sin(submodphas + Pphas) + \
                  Tot_D_iR_amp[select] * Tot_D_jL_amp[select] * Pamp * submodamp * np.sin(submodphas + Tot_D_iR_phas[select] - Tot_D_jL_phas[select] - Pphas + 2. * (pang1[select] + pang2[select]))
        
                LiRj_Real[select] += Pamp * submodamp * np.cos(submodphas - Pphas) + \
                  Tot_D_iL_amp[select] * Tot_D_jR_amp[select] * Pamp * submodamp * np.cos(submodphas + Tot_D_iL_phas[select] - Tot_D_jR_phas[select] + Pphas - 2. * (pang1[select] + pang2[select]))
        
                LiRj_Imag[select] += Pamp * submodamp * np.sin(submodphas - Pphas) + \
                  Tot_D_iL_amp[select] * Tot_D_jR_amp[select] * Pamp * submodamp * np.sin(submodphas + Tot_D_iL_phas[select] - Tot_D_jR_phas[select] + Pphas - 2. * (pang1[select] + pang2[select]))     
    
    
    RiLj_Real += \
      Tot_D_iR_amp * llamp * np.cos(Tot_D_iR_phas + llphas + 2. * pang1) + \
      Tot_D_jL_amp * rramp * np.cos(-Tot_D_jL_phas + rrphas + 2. * pang2)
    
    RiLj_Imag += \
      Tot_D_iR_amp * llamp * np.sin(Tot_D_iR_phas + llphas + 2. * pang1) + \
      Tot_D_jL_amp * rramp * np.sin(-Tot_D_jL_phas + rrphas + 2. * pang2)

    LiRj_Real += \
      Tot_D_iL_amp * rramp * np.cos(Tot_D_iL_phas + rrphas - 2. * pang1) + \
      Tot_D_jR_amp * llamp * np.cos(-Tot_D_jR_phas + llphas - 2. * pang2)

    LiRj_Imag += \
      Tot_D_iL_amp * rramp * np.sin(Tot_D_iL_phas + rrphas - 2. * pang1) + \
      Tot_D_jR_amp * llamp * np.sin(-Tot_D_jR_phas + llphas - 2. * pang2)
   

    compv = np.concatenate([LiRj_Real, LiRj_Imag, RiLj_Real, RiLj_Imag])
        
    return compv




def synthetic_deq(nant, pang1, pang2, ant1, ant2, llamp, llphas, rramp, rrphas, model_ireal, model_iimag, model_rlreal, model_rlimag, model_lrreal, model_lrimag, stokes, *p):
    
    RiRj_Real = np.zeros(len(pang1))
    RiRj_Imag = np.zeros(len(pang1))
    LiLj_Real = np.zeros(len(pang1))
    LiLj_Imag = np.zeros(len(pang1))
    
    RiLj_Real = np.zeros(len(pang1))
    RiLj_Imag = np.zeros(len(pang1))
    LiRj_Real = np.zeros(len(pang1))
    LiRj_Imag = np.zeros(len(pang1))
    
    model_iamp = np.abs(model_ireal + 1j * model_iimag)
    model_iphas = np.angle(model_ireal + 1j * model_iimag)
    model_rlamp = np.abs(model_rlreal + 1j * model_rlimag)
    model_rlphas = np.angle(model_rlreal + 1j * model_rlimag)
    model_lramp = np.abs(model_lrreal + 1j * model_lrimag)
    model_lrphas = np.angle(model_lrreal + 1j * model_lrimag)
    
    dump = np.array(p)
    
    dumreal = dump[2*ant1]
    dumimag = dump[2*ant1 + 1]
    Tot_D_iR_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_iR_phas = np.angle(dumreal + 1j*dumimag)

    dumreal = dump[2*nant + 2*ant2]
    dumimag = dump[2*nant + 2*ant2 + 1]
    Tot_D_jL_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_jL_phas = np.angle(dumreal + 1j*dumimag)
    
    dumreal = dump[2*nant + 2*ant1]
    dumimag = dump[2*nant + 2*ant1 + 1]
    Tot_D_iL_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_iL_phas = np.angle(dumreal + 1j*dumimag)
    
    dumreal = dump[2*ant2]
    dumimag = dump[2*ant2 + 1]
    Tot_D_jR_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_jR_phas = np.angle(dumreal + 1j*dumimag)
    
    
    RiRj_Real += model_ireal + Tot_D_iR_amp * Tot_D_jR_amp * model_iamp * np.cos(model_iphas + Tot_D_iR_phas - Tot_D_jR_phas + 2. * (pang1 - pang2))
    RiRj_Imag += model_iimag + Tot_D_iR_amp * Tot_D_jR_amp * model_iamp * np.sin(model_iphas + Tot_D_iR_phas - Tot_D_jR_phas + 2. * (pang1 - pang2))
    LiLj_Real += model_ireal + Tot_D_iL_amp * Tot_D_jL_amp * model_iamp * np.cos(model_iphas + Tot_D_iL_phas - Tot_D_jL_phas - 2. * (pang1 - pang2))
    LiLj_Imag += model_iimag + Tot_D_iL_amp * Tot_D_jL_amp * model_iamp * np.sin(model_iphas + Tot_D_iL_phas - Tot_D_jL_phas - 2. * (pang1 - pang2))
    
    RiRj_Real += \
        Tot_D_iR_amp * model_lramp * np.cos(Tot_D_iR_phas + model_lrphas + 2. * pang1) + \
        Tot_D_jR_amp * model_rlamp * np.cos(-Tot_D_jR_phas + model_rlphas - 2. * pang2)
    
    RiRj_Imag += \
        Tot_D_iR_amp * model_lramp * np.sin(Tot_D_iR_phas + model_lrphas + 2. * pang1) + \
        Tot_D_jR_amp * model_rlamp * np.sin(-Tot_D_jR_phas + model_rlphas - 2. * pang2)
    
    LiLj_Real += \
        Tot_D_iL_amp * model_rlamp * np.cos(Tot_D_iL_phas + model_rlphas - 2. * pang1) + \
        Tot_D_jL_amp * model_lramp * np.cos(-Tot_D_jL_phas + model_lrphas + 2. * pang2)
    
    LiLj_Imag += \
        Tot_D_iL_amp * model_rlamp * np.sin(Tot_D_iL_phas + model_rlphas - 2. * pang1) + \
        Tot_D_jL_amp * model_lramp * np.sin(-Tot_D_jL_phas + model_lrphas + 2. * pang2)
        
    
    RiLj_Real += model_rlreal + Tot_D_iR_amp * Tot_D_jL_amp * model_lramp * np.cos(model_lrphas + Tot_D_iR_phas - Tot_D_jL_phas + 2. * (pang1 + pang2))
    RiLj_Imag += model_rlimag + Tot_D_iR_amp * Tot_D_jL_amp * model_lramp * np.sin(model_lrphas + Tot_D_iR_phas - Tot_D_jL_phas + 2. * (pang1 + pang2))
    LiRj_Real += model_lrreal + Tot_D_iL_amp * Tot_D_jR_amp * model_rlamp * np.cos(model_rlphas + Tot_D_iL_phas - Tot_D_jR_phas - 2. * (pang1 + pang2))
    LiRj_Imag += model_lrimag + Tot_D_iL_amp * Tot_D_jR_amp * model_rlamp * np.sin(model_rlphas + Tot_D_iL_phas - Tot_D_jR_phas - 2. * (pang1 + pang2))
        
    RiLj_Real += \
      Tot_D_iR_amp * model_iamp * np.cos(Tot_D_iR_phas + model_iphas + 2. * pang1) + \
      Tot_D_jL_amp * model_iamp * np.cos(-Tot_D_jL_phas + model_iphas + 2. * pang2)
    
    RiLj_Imag += \
      Tot_D_iR_amp * model_iamp * np.sin(Tot_D_iR_phas + model_iphas + 2. * pang1) + \
      Tot_D_jL_amp * model_iamp * np.sin(-Tot_D_jL_phas + model_iphas + 2. * pang2)

    LiRj_Real += \
      Tot_D_iL_amp * model_iamp * np.cos(Tot_D_iL_phas + model_iphas - 2. * pang1) + \
      Tot_D_jR_amp * model_iamp * np.cos(-Tot_D_jR_phas + model_iphas - 2. * pang2)

    LiRj_Imag += \
      Tot_D_iL_amp * model_iamp * np.sin(Tot_D_iL_phas + model_iphas - 2. * pang1) + \
      Tot_D_jR_amp * model_iamp * np.sin(-Tot_D_jR_phas + model_iphas - 2. * pang2)  
   
    if(stokes == 'RR'): return RiRj_Real + 1j * RiRj_Imag
    if(stokes == 'LL'): return LiLj_Real + 1j * LiLj_Imag
    if(stokes == 'LR'): return LiRj_Real + 1j * LiRj_Imag
    if(stokes == 'RL'): return RiLj_Real + 1j * RiLj_Imag



def pol_deq(x, *p):
    """
    The D-term models for the D-term estimation with instrumental polarization self-calibration.
    
    Args:
        x: dummy parameters (not to be used).
        *p (args): the best-fit parameters args.
    """
    
    nant, pang1, pang2, ant1, ant2, rramp, rrphas, llamp, llphas, model_rlreal, model_rlimag, model_rlamp, model_rlphas, model_lrreal, model_lrimag, model_lramp, model_lrphas = inputz
    
    RiLj_Real = np.zeros(len(pang1))
    RiLj_Imag = np.zeros(len(pang1))
    LiRj_Real = np.zeros(len(pang1))
    LiRj_Imag = np.zeros(len(pang1))
    
    
    dump = np.array(p)
    
    dumreal = dump[2*ant1]
    dumimag = dump[2*ant1 + 1]
    Tot_D_iR_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_iR_phas = np.angle(dumreal + 1j*dumimag)

    dumreal = dump[2*nant + 2*ant2]
    dumimag = dump[2*nant + 2*ant2 + 1]
    Tot_D_jL_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_jL_phas = np.angle(dumreal + 1j*dumimag)
    
    dumreal = dump[2*nant + 2*ant1]
    dumimag = dump[2*nant + 2*ant1 + 1]
    Tot_D_iL_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_iL_phas = np.angle(dumreal + 1j*dumimag)
    
    dumreal = dump[2*ant2]
    dumimag = dump[2*ant2 + 1]
    Tot_D_jR_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_jR_phas = np.angle(dumreal + 1j*dumimag)
    
    
    RiLj_Real += model_rlreal + \
      Tot_D_iR_amp * Tot_D_jL_amp * model_lramp * np.cos(model_lrphas + Tot_D_iR_phas - Tot_D_jL_phas + 2. * (pang1 + pang2))

    RiLj_Imag += model_rlimag + \
      Tot_D_iR_amp * Tot_D_jL_amp * model_lramp * np.sin(model_lrphas + Tot_D_iR_phas - Tot_D_jL_phas + 2. * (pang1 + pang2))

    LiRj_Real += model_lrreal + \
      Tot_D_iL_amp * Tot_D_jR_amp * model_rlamp * np.cos(model_rlphas + Tot_D_iL_phas - Tot_D_jR_phas - 2. * (pang1 + pang2))

    LiRj_Imag += model_lrimag + \
      Tot_D_iL_amp * Tot_D_jR_amp * model_rlamp * np.sin(model_rlphas + Tot_D_iL_phas - Tot_D_jR_phas - 2. * (pang1 + pang2))
    
    
    RiLj_Real += \
      Tot_D_iR_amp * llamp * np.cos(Tot_D_iR_phas + llphas + 2. * pang1) + \
      Tot_D_jL_amp * rramp * np.cos(-Tot_D_jL_phas + rrphas + 2. * pang2)
    
    RiLj_Imag += \
      Tot_D_iR_amp * llamp * np.sin(Tot_D_iR_phas + llphas + 2. * pang1) + \
      Tot_D_jL_amp * rramp * np.sin(-Tot_D_jL_phas + rrphas + 2. * pang2)

    LiRj_Real += \
      Tot_D_iL_amp * rramp * np.cos(Tot_D_iL_phas + rrphas - 2. * pang1) + \
      Tot_D_jR_amp * llamp * np.cos(-Tot_D_jR_phas + llphas - 2. * pang2)

    LiRj_Imag += \
      Tot_D_iL_amp * rramp * np.sin(Tot_D_iL_phas + rrphas - 2. * pang1) + \
      Tot_D_jR_amp * llamp * np.sin(-Tot_D_jR_phas + llphas - 2. * pang2)  
   

    compv = np.concatenate([LiRj_Real, LiRj_Imag, RiLj_Real, RiLj_Imag])
    
    
    return compv



def pol_deq_comp(comp, parmset, *p):
    """
    Extract component-wise best-fit models with instrumental polarization self-calibration.
    
    Args:
        comp (str): the name of the component to be extracted.
        *p (args): the best-fit parameters args.
    """
    
    (nant, polcalsour, sourcearr, pang1, pang2, ant1, ant2, model_rlreal, model_rlimag, model_lrreal, model_lrimag, rramp, rrphas, llamp, llphas) = parmset
    
    model_rlamp = np.abs(model_rlreal + 1j * model_rlimag)
    model_rlphas = np.angle(model_rlreal + 1j * model_rlimag)
    model_lramp = np.abs(model_lrreal + 1j * model_lrimag)
    model_lrphas = np.angle(model_lrreal + 1j * model_lrimag)
    
    
    pol_RiLj_Real = np.zeros(len(pang1))
    pol_RiLj_Imag = np.zeros(len(pang1))
    pol_LiRj_Real = np.zeros(len(pang1))
    pol_LiRj_Imag = np.zeros(len(pang1))
    
    dterm_RiLj_Real = np.zeros(len(pang1))
    dterm_RiLj_Imag = np.zeros(len(pang1))
    dterm_LiRj_Real = np.zeros(len(pang1))
    dterm_LiRj_Imag = np.zeros(len(pang1))
    
    second_RiLj_Real = np.zeros(len(pang1))
    second_RiLj_Imag = np.zeros(len(pang1))
    second_LiRj_Real = np.zeros(len(pang1))
    second_LiRj_Imag = np.zeros(len(pang1))
    
    
    dump = np.array(p)
    
    
    dumreal = np.array([dump[2*int(s)] for s in ant1])
    dumimag = np.array([dump[2*int(s)+1] for s in ant1])
    Tot_D_iR_amp = np.sqrt(dumreal**2 + dumimag**2)
    Tot_D_iR_phas = np.angle(dumreal + 1j*dumimag)

    dumreal = np.array([dump[2*nant + 2*int(s)] for s in ant2])
    dumimag = np.array([dump[2*nant + 2*int(s)+1] for s in ant2])
    Tot_D_jL_amp = np.sqrt(dumreal**2 + dumimag**2)
    Tot_D_jL_phas = np.angle(dumreal + 1j*dumimag)


    dumreal = np.array([dump[2*nant + 2*int(s)] for s in ant1])
    dumimag = np.array([dump[2*nant + 2*int(s)+1] for s in ant1])
    Tot_D_iL_amp = np.sqrt(dumreal**2 + dumimag**2)
    Tot_D_iL_phas = np.angle(dumreal + 1j*dumimag)


    dumreal = np.array([dump[2*int(s)] for s in ant2])
    dumimag = np.array([dump[2*int(s)+1] for s in ant2])
    Tot_D_jR_amp = np.sqrt(dumreal**2 + dumimag**2)
    Tot_D_jR_phas = np.angle(dumreal + 1j*dumimag)
    

    for l in range(len(polcalsour)):
      
        select = (sourcearr == polcalsour[l])
        
        pol_RiLj_Real[select] += model_rlreal[select]
        pol_RiLj_Imag[select] += model_rlimag[select]
        pol_LiRj_Real[select] += model_lrreal[select]
        pol_LiRj_Imag[select] += model_lrimag[select]
        
        second_RiLj_Real[select] += Tot_D_iR_amp[select] * Tot_D_jL_amp[select] * model_rlamp[select] * np.cos(model_rlphas[select] + Tot_D_iR_phas[select] - Tot_D_jL_phas[select] + 2. * (pang1[select] + pang2[select]))
        second_RiLj_Imag[select] += Tot_D_iR_amp[select] * Tot_D_jL_amp[select] * model_rlamp[select] * np.sin(model_rlphas[select] + Tot_D_iR_phas[select] - Tot_D_jL_phas[select] + 2. * (pang1[select] + pang2[select]))
        second_LiRj_Real[select] += Tot_D_iL_amp[select] * Tot_D_jR_amp[select] * model_lramp[select] * np.cos(model_lrphas[select] + Tot_D_iL_phas[select] - Tot_D_jR_phas[select] - 2. * (pang1[select] + pang2[select]))
        second_LiRj_Imag[select] += Tot_D_iL_amp[select] * Tot_D_jR_amp[select] * model_lramp[select] * np.sin(model_lrphas[select] + Tot_D_iL_phas[select] - Tot_D_jR_phas[select] - 2. * (pang1[select] + pang2[select]))
          
        
    dterm_RiLj_Real += Tot_D_iR_amp * llamp * np.cos(Tot_D_iR_phas + llphas + 2. * pang1) + Tot_D_jL_amp * rramp * np.cos(-Tot_D_jL_phas + rrphas + 2. * pang2)
    dterm_RiLj_Imag += Tot_D_iR_amp * llamp * np.sin(Tot_D_iR_phas + llphas + 2. * pang1) + Tot_D_jL_amp * rramp * np.sin(-Tot_D_jL_phas + rrphas + 2. * pang2)
    dterm_LiRj_Real += Tot_D_iL_amp * rramp * np.cos(Tot_D_iL_phas + rrphas - 2. * pang1) + Tot_D_jR_amp * llamp * np.cos(-Tot_D_jR_phas + llphas - 2. * pang2)
    dterm_LiRj_Imag += Tot_D_iL_amp * rramp * np.sin(Tot_D_iL_phas + rrphas - 2. * pang1) + Tot_D_jR_amp * llamp * np.sin(-Tot_D_jR_phas + llphas - 2. * pang2)  


    pol_rl, pol_lr = pol_RiLj_Real + 1j * pol_RiLj_Imag, pol_LiRj_Real + 1j * pol_LiRj_Imag
    pol_q, pol_u = (pol_rl + pol_lr) / 2., -1j * (pol_rl - pol_lr) / 2.
    pol_qamp, pol_qphas, pol_uamp, pol_uphas = np.absolute(pol_q), np.angle(pol_q), np.absolute(pol_u), np.angle(pol_u)
    
    dterm_rl, dterm_lr = dterm_RiLj_Real + 1j * dterm_RiLj_Imag, dterm_LiRj_Real + 1j * dterm_LiRj_Imag
    dterm_q, dterm_u = (dterm_rl + dterm_lr) / 2., -1j * (dterm_rl - dterm_lr) / 2.
    dterm_qamp, dterm_qphas, dterm_uamp, dterm_uphas = np.absolute(dterm_q), np.angle(dterm_q), np.absolute(dterm_u), np.angle(dterm_u)
    
    second_rl, second_lr = second_RiLj_Real + 1j * second_RiLj_Imag, second_LiRj_Real + 1j * second_LiRj_Imag
    second_q, second_u = (second_rl + second_lr) / 2., -1j * (second_rl - second_lr) / 2.
    second_qamp, second_qphas, second_uamp, second_uphas = np.absolute(second_q), np.angle(second_q), np.absolute(second_u), np.angle(second_u)


    if(comp == 'pol'): return pol_qamp, pol_qphas, pol_uamp, pol_uphas
    if(comp == 'dterm'): return dterm_qamp, dterm_qphas, dterm_uamp, dterm_uphas
    if(comp == 'second'): return second_qamp, second_qphas, second_uamp, second_uphas



def pol_deq_inputz(inputz, *p):
    """
    The D-term models for the D-term estimation with instrumental polarization self-calibration.
    
    Args:
        x: dummy parameters (not to be used).
        *p (args): the best-fit parameters args.
    """
    
    nant, pang1, pang2, ant1, ant2, rramp, rrphas, llamp, llphas, model_rlreal, model_rlimag, model_rlamp, model_rlphas, model_lrreal, model_lrimag, model_lramp, model_lrphas = inputz
    
    RiLj_Real = np.zeros(len(pang1))
    RiLj_Imag = np.zeros(len(pang1))
    LiRj_Real = np.zeros(len(pang1))
    LiRj_Imag = np.zeros(len(pang1))
    
    
    dump = np.array(p)
    
    dumreal = dump[2*ant1]
    dumimag = dump[2*ant1 + 1]
    Tot_D_iR_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_iR_phas = np.angle(dumreal + 1j*dumimag)

    dumreal = dump[2*nant + 2*ant2]
    dumimag = dump[2*nant + 2*ant2 + 1]
    Tot_D_jL_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_jL_phas = np.angle(dumreal + 1j*dumimag)
    
    dumreal = dump[2*nant + 2*ant1]
    dumimag = dump[2*nant + 2*ant1 + 1]
    Tot_D_iL_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_iL_phas = np.angle(dumreal + 1j*dumimag)
    
    dumreal = dump[2*ant2]
    dumimag = dump[2*ant2 + 1]
    Tot_D_jR_amp = np.absolute(dumreal + 1j * dumimag)
    Tot_D_jR_phas = np.angle(dumreal + 1j*dumimag)
    

    RiLj_Real += model_rlreal + \
      Tot_D_iR_amp * Tot_D_jL_amp * model_lramp * np.cos(model_lrphas + Tot_D_iR_phas - Tot_D_jL_phas + 2. * (pang1 + pang2))

    RiLj_Imag += model_rlimag + \
      Tot_D_iR_amp * Tot_D_jL_amp * model_lramp * np.sin(model_lrphas + Tot_D_iR_phas - Tot_D_jL_phas + 2. * (pang1 + pang2))

    LiRj_Real += model_lrreal + \
      Tot_D_iL_amp * Tot_D_jR_amp * model_rlamp * np.cos(model_rlphas + Tot_D_iL_phas - Tot_D_jR_phas - 2. * (pang1 + pang2))

    LiRj_Imag += model_lrimag + \
      Tot_D_iL_amp * Tot_D_jR_amp * model_rlamp * np.sin(model_rlphas + Tot_D_iL_phas - Tot_D_jR_phas - 2. * (pang1 + pang2))
    
        
    RiLj_Real += \
      Tot_D_iR_amp * llamp * np.cos(Tot_D_iR_phas + llphas + 2. * pang1) + \
      Tot_D_jL_amp * rramp * np.cos(-Tot_D_jL_phas + rrphas + 2. * pang2)
    
    RiLj_Imag += \
      Tot_D_iR_amp * llamp * np.sin(Tot_D_iR_phas + llphas + 2. * pang1) + \
      Tot_D_jL_amp * rramp * np.sin(-Tot_D_jL_phas + rrphas + 2. * pang2)

    LiRj_Real += \
      Tot_D_iL_amp * rramp * np.cos(Tot_D_iL_phas + rrphas - 2. * pang1) + \
      Tot_D_jR_amp * llamp * np.cos(-Tot_D_jR_phas + llphas - 2. * pang2)

    LiRj_Imag += \
      Tot_D_iL_amp * rramp * np.sin(Tot_D_iL_phas + rrphas - 2. * pang1) + \
      Tot_D_jR_amp * llamp * np.sin(-Tot_D_jR_phas + llphas - 2. * pang2)  
   

    compv = np.concatenate([LiRj_Real, LiRj_Imag, RiLj_Real, RiLj_Imag])
    
    return compv


def rescale_weight(ant1, ant2, inputsigma, antname, weightfactors, logger):
    
    nant = len(antname)
    
    sigmaant1 = np.concatenate([ant1, ant1, ant1, ant1])
    sigmaant2 = np.concatenate([ant2, ant2, ant2, ant2])
    
    outputweight = 1. / inputsigma ** 2
    for m in range(nant):
        if(antname[m] in weightfactors):
            outputweight[(sigmaant1 == m) | (sigmaant2 == m)] = outputweight[(sigmaant1 == m) | (sigmaant2 == m)] * weightfactors.get(antname[m])
            logger.info('The visibility weights for {:s} station are rescaled by a factor of {:4.2f}.'.format(antname[m], weightfactors.get(antname[m])))
        
    logger.info(' ')
    
    outputsigma = 1. / outputweight ** (1. / 2.)
    
    return outputsigma


def arrange_index(orig_nant, ant1, ant2, logger, antname, printmessage = True):

    nant = np.copy(orig_nant)
    
    dum = 0
    
    removed_Index = []
    
    for m in range(orig_nant):
        if((sum(ant1 == dum) == 0) & (sum(ant2 == dum) == 0)):
            ant1[ant1 > dum] -= 1
            ant2[ant2 > dum] -= 1
            nant -= 1
            removed_Index.append(m)
            if printmessage:
                logger.info('{:s} has no data, the fitting will not solve the D-Terms for it.'.format(antname[m]))
        else:
            dum += 1
    
    if(len(removed_Index) != 0): antname = np.delete(antname, removed_Index)
    
    return nant, antname, removed_Index, ant1, ant2
    
        


def pol_gpcal(direc, outputname, data, calsour, antname, logger, filetype = 'png', printmessage = True, Dbound = 1.0, Pbound = 1.0, manualweight = False, weightfactors = None, fixdterm = False, fixdr = None, fixdl = None, \
              multiproc = True, nproc = 2, colors = None, vplot = False, allplot = False, vplot_title = None, vplot_scanavg = False, vplot_avg_nat = False, tsep = None, \
              resplot = False, markerarr = None):
    
    """
    Estimate the D-terms using instrumental polarization self-calibration.
    
    Args:
        spoliter (int): the number of iteration of instrumental polarization self-calibration.
    """  

    result_DRArr, result_DLArr, result_ant = [], [], []
          
    time, dayarr, sourcearr, pang1, pang2, ant1, ant2, rlreal, rlimag, lrreal, lrimag, rlsigma, lrsigma, rramp, rrphas, llamp, llphas, \
    model_rlreal, model_rlimag, model_lrreal, model_lrimag, model_rlamp, model_rlphas, model_lramp, model_lrphas = \
        np.array(data["time"]), np.array(data["day"]), np.array(data["source"]), np.array(data["pang1"]), np.array(data["pang2"]), np.array(data["ant1"]), np.array(data["ant2"]), \
        np.array(data["rlreal"]), np.array(data["rlimag"]), np.array(data["lrreal"]), np.array(data["lrimag"]), \
        np.array(data["rlsigma"]), np.array(data["lrsigma"]), np.array(data["rramp"]), np.array(data["rrphas"]), np.array(data["llamp"]), np.array(data["llphas"]), \
        np.array(data["model_rlreal"]), np.array(data["model_rlimag"]), np.array(data["model_lrreal"]), np.array(data["model_lrimag"]), \
        np.array(data["model_rlamp"]), np.array(data["model_rlphas"]), np.array(data["model_lramp"]), np.array(data["model_lrphas"])
    
    qamp = np.array(data["qamp"])
    qphas = np.array(data["qphas"])
    qsigma = np.array(data["qsigma"])
    uamp = np.array(data["uamp"])
    uphas = np.array(data["uphas"])
    usigma = np.array(data["usigma"])
    
    nant = len(antname)


    inputx = np.concatenate([pang1, pang1, pang1, pang1])
    inputy = np.concatenate([lrreal, lrimag, rlreal, rlimag])
    inputsigma = np.concatenate([lrsigma, lrsigma, rlsigma, rlsigma])
    
    # Rescale the visibility weights of specific stations if requested.
    if manualweight:
        outputsigma = rescale_weight(ant1, ant2, inputsigma, antname, weightfactors, logger)
    else:
        outputsigma = np.copy(inputsigma)
        if printmessage:
            logger.info('No visibility weight rescaling applied.\n')
    
    
    dumantname = np.copy(antname)
    orig_ant1 = np.copy(ant1)
    orig_ant2 = np.copy(ant2)
    
    orig_nant = np.copy(nant)
    
    
    nant, dumantname, removed_Index, ant1, ant2 = arrange_index(orig_nant, ant1, ant2, logger, antname, printmessage = printmessage)
    
    init = np.zeros(2*2*nant)
    
    # The boundaries of parameters allowed for the least-square fitting.
    lbound = [-Dbound]*(2*2*nant)
    ubound = [Dbound]*(2*2*nant)
    

    if fixdterm:
        for i in range(nant):
            if dumantname[i] in fixdr:
                lbound[2*i] = np.real(fixdr.get(dumantname[i])) - 1e-8
                ubound[2*i] = np.real(fixdr.get(dumantname[i])) + 1e-8
                lbound[2*i+1] = np.imag(fixdr.get(dumantname[i])) - 1e-8
                ubound[2*i+1] = np.imag(fixdr.get(dumantname[i])) + 1e-8
                lbound[2*nant+2*i] = np.real(fixdl.get(dumantname[i])) - 1e-8
                ubound[2*nant+2*i] = np.real(fixdl.get(dumantname[i])) + 1e-8
                lbound[2*nant+2*i+1] = np.imag(fixdl.get(dumantname[i])) - 1e-8
                ubound[2*nant+2*i+1] = np.imag(fixdl.get(dumantname[i])) + 1e-8
                init[2*i] = np.real(fixdr.get(dumantname[i]))
                init[2*i+1] = np.imag(fixdr.get(dumantname[i]))
                init[2*nant+2*i] = np.real(fixdl.get(dumantname[i]))
                init[2*nant+2*i+1] = np.imag(fixdl.get(dumantname[i]))
    
    
    bounds=(lbound,ubound)
    
    
    global inputz
    
    inputz = (nant, pang1, pang2, ant1, ant2, rramp, rrphas, llamp, llphas, model_rlreal, model_rlimag, model_rlamp, model_rlphas, model_lrreal, model_lrimag, model_lramp, model_lrphas)
    
    
    Iteration, pco = curve_fit(pol_deq, inputx, inputy, p0=init, sigma = outputsigma, absolute_sigma = False, bounds = bounds)
        
    error = np.sqrt(np.diag(pco))


    # Restore the original antenna numbers.
    insert_index = []
    
    dum = 0
    for it in removed_Index:
        insert_index.append(2*it - 2*dum)
        insert_index.append(2*it - 2*dum)
        insert_index.append(2*nant + 2*it - 2*dum)
        insert_index.append(2*nant + 2*it - 2*dum)
        dum += 1
        
    Iteration = np.insert(Iteration, insert_index, [0.]*len(insert_index))
    error = np.insert(error, insert_index, [0.]*len(insert_index))
    
    # logger.info('The fitting is completed within {:d} seconds.\n'.format(int(round(time2 - time1))))
    
    
    nant = np.copy(orig_nant)
    
    ant1 = np.copy(orig_ant1)
    ant2 = np.copy(orig_ant2)
    
    
    for m in range(nant):
        result_ant.append(antname[m])
        result_DRArr.append(Iteration[2*m] + 1j*Iteration[2*m+1])
        result_DLArr.append(Iteration[2*nant+2*m] + 1j*Iteration[2*nant+2*m+1])


    
    inputz = (nant, pang1, pang2, ant1, ant2, rramp, rrphas, llamp, llphas, model_rlreal, model_rlimag, model_rlamp, model_rlphas, model_lrreal, model_lrimag, model_lramp, model_lrphas)
    
    # Calculate the reduced chi-square of the fitting.
    dumfit = pol_deq(inputx, *Iteration)

    ydata = np.concatenate([lrreal+1j*lrimag, rlreal+1j*rlimag])
    yfit = np.concatenate([dumfit[0:len(lrreal)]+1j*dumfit[len(lrreal):len(lrreal)*2], dumfit[len(lrreal)*2:len(lrreal)*3]+1j*dumfit[len(lrreal)*3:len(lrreal)*4]])
    ysigma = np.concatenate([lrsigma, rlsigma])
    
    chisq_num = np.sum(np.abs(((ydata - yfit) / ysigma) ** 2))
    chisq_den = (2. * len(ydata))
    # chisq_den = (2. * len(ydata) - float(len(Iteration)))
    
    chisq = chisq_num / chisq_den
    
    
    del inputz
    
    if printmessage:
        logger.info('The reduced chi-square of the fitting is {:5.3f}.\n'.format(chisq))
    
        for m in range(nant):
            logger.info('{:s}: RCP - amplitude = {:6.3f} %, phase = {:7.2f} deg, LCP - amplitude = {:6.3f} %, phase = {:7.2f} deg'.format(antname[m], \
              np.abs(Iteration[2*m] + 1j*Iteration[2*m+1]) * 1e2, np.degrees(np.angle((Iteration[2*m] + 1j*Iteration[2*m+1]))), \
              np.abs(Iteration[2*nant+2*m] + 1j*Iteration[2*nant+2*m+1]) * 1e2, np.degrees(np.angle(Iteration[2*nant+2*m] + 1j*Iteration[2*nant+2*m+1]))))
    
        logger.info(' ')
    
    
    mod_lr, mod_rl = dumfit[0:len(rlreal)] + 1j*dumfit[len(rlreal):len(rlreal)*2], dumfit[len(rlreal)*2:len(rlreal)*3] + 1j*dumfit[len(rlreal)*3:len(rlreal)*4]
    mod_q, mod_u = (mod_rl + mod_lr) / 2., -1j * (mod_rl - mod_lr) / 2.
    mod_qamp, mod_qphas, mod_uamp, mod_uphas = np.absolute(mod_q), np.angle(mod_q), np.absolute(mod_u), np.angle(mod_u)
    
    
        
    parmset = (nant, calsour, sourcearr, pang1, pang2, ant1, ant2, model_rlreal, model_rlimag, model_lrreal, model_lrimag, rramp, rrphas, llamp, llphas)
        
    if allplot:
        mod_pol_qamp, mod_pol_qphas, mod_pol_uamp, mod_pol_uphas = pol_deq_comp('pol', parmset, *Iteration)
        mod_dterm_qamp, mod_dterm_qphas, mod_dterm_uamp, mod_dterm_uphas = pol_deq_comp('dterm', parmset, *Iteration)
        mod_second_qamp, mod_second_qphas, mod_second_uamp, mod_second_uphas = pol_deq_comp('second', parmset, *Iteration)
        allplots = (mod_pol_qamp, mod_pol_qphas, mod_pol_uamp, mod_pol_uphas, mod_dterm_qamp, mod_dterm_qphas, mod_dterm_uamp, mod_dterm_uphas)
    else:
        allplots = None
    
    # Create vplots if requested.
    if vplot:
        
        logger.info('\nCreating vplots for all baselines... It may take some time.')
        
        ifn = np.unique(data.loc[:, "IF"])[0]
        
        parmset = []

        for l in range(len(calsour)):
            for m in range(nant):
                for n in range(nant):
                    if(m==n):
                        continue
                    dumm = m
                    dumn = n
                    if(m>n):
                        dumm = n
                        dumn = m
                        
                        select = (ant1 == dumm) & (ant2 == dumn) & (sourcearr == calsour[l])
                    
                        selected_time, selected_day, selected_qamp, selected_qphas, selected_qsigma, selected_uamp, selected_uphas, selected_usigma, selected_mod_qamp, selected_mod_qphas, selected_mod_uamp, selected_mod_uphas = \
                            time[select], dayarr[select], qamp[select], qphas[select], qsigma[select], uamp[select], uphas[select], usigma[select], \
                            mod_qamp[select], mod_qphas[select], mod_uamp[select], mod_uphas[select]
                            
                        if allplot:
                            selected_mod_pol_qamp = mod_pol_qamp[select]
                            selected_mod_pol_qphas = mod_pol_qphas[select]
                            selected_mod_pol_uamp = mod_pol_uamp[select]
                            selected_mod_pol_uphas = mod_pol_uphas[select]
                            
                            selected_mod_dterm_qamp = mod_dterm_qamp[select]
                            selected_mod_dterm_qphas = mod_dterm_qphas[select]
                            selected_mod_dterm_uamp = mod_dterm_uamp[select]
                            selected_mod_dterm_uphas = mod_dterm_uphas[select]
                            
                            allplots = (selected_mod_pol_qamp, selected_mod_pol_qphas, selected_mod_pol_uamp, selected_mod_pol_uphas, \
                                        selected_mod_dterm_qamp, selected_mod_dterm_qphas, selected_mod_dterm_uamp, selected_mod_dterm_uphas)
                        
                            
                        if multiproc:
                            parmset.append((colors[l], calsour[l], ifn, antname[dumm], antname[dumn], \
                                            selected_day, selected_time, selected_qamp, selected_qphas, selected_qsigma, selected_uamp, selected_uphas, selected_usigma, \
                                            selected_mod_qamp, selected_mod_qphas, selected_mod_uamp, selected_mod_uphas, direc+'gpcal/'+outputname+'pol.vplot.IF'+str(ifn), filetype, \
                                            allplots, vplot_title, vplot_scanavg, vplot_avg_nat, tsep))
                        else:
                            ph.visualplot(calsour[l], ifn, antname[dumm], antname[dumn], \
                                            selected_day, selected_time, selected_qamp, selected_qphas, selected_qsigma, selected_uamp, selected_uphas, selected_usigma, \
                                            selected_mod_qamp, selected_mod_qphas, selected_mod_uamp, selected_mod_uphas, direc+'gpcal/'+outputname+'pol.vplot.IF'+str(ifn), filetype, \
                                            allplots = allplots, title = vplot_title, scanavg = vplot_scanavg, avg_nat = vplot_avg_nat, tsep = tsep, color = colors[l])


        if multiproc:
            ph.visualplot_run(parmset, nproc = nproc)

        logger.info('...completed.\n')
    
    if resplot:
        ph.residualplot(markerarr, colors, dayarr, ifn-1, nant, antname, calsour, \
                        dumfit, time, ant1, ant2, sourcearr, qamp, qphas, uamp, uphas, qsigma, usigma, direc+'gpcal/'+outputname+'pol.resplot.IF'+str(ifn), tsep = 2. / 60., title = outputname, filetype = filetype)
    
    result_ant, result_DRArr, result_DLArr = np.array(result_ant), np.array(result_DRArr), np.array(result_DLArr)
    
    return (Iteration, error), (result_ant, result_DRArr, result_DLArr), (chisq_num, chisq_den, chisq)






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 29 13:18:58 2021

@author: jpark
"""


import numpy as np
import pandas as pd
import os, sys
from os import path

import timeit
import logging
import copy

from AIPS import AIPS
from AIPSTask import AIPSTask
from AIPSData import AIPSUVData, AIPSImage
from Wizardry.AIPSData import AIPSUVData as WAIPSUVData


# Suppress AIPS printing on terminal
def blockPrint():
    sys.stdout = open(os.devnull, 'w')

# Enable AIPS printing on terminal
def enablePrint():
    sys.stdout = sys.__stdout__
    
    
def runfitld(outname, outclass, datain):
    """
    Run FITLD in AIPS.
    
    Args:
        outname (str): a FITLD parameter outname.
        outclass (str): a FITLD parameter outclass.
        datain (str): a FITLD parameter datain.
    """
    
    data = AIPSUVData(outname, outclass, 1, 1)
    if(data.exists() == True):
        data.clrstat()
        data.zap()
    
    image = AIPSImage(outname, outclass, 1, 1)
    if(image.exists() == True):
        image.clrstat()
        image.zap()
        
    fitld = AIPSTask('fitld')
    fitld.outname = outname
    fitld.outclass = outclass
    dumdatain = datain.split('/')[-1]
    if(len(dumdatain) >= 46):
        newdatain = "/".join(datain.split('/')[:-1]) + '/' + dumdatain[0:45]
        os.system('cp {:} {:}'.format(datain, newdatain))
        fitld.datain = newdatain
    else:
        fitld.datain = datain

    fitld.outseq = 1
    blockPrint()
    fitld.go()    
    enablePrint()
    if(len(dumdatain) >= 46):
        os.system('rm {:}'.format(newdatain))


def runlpcal(inname, inclass, in2class, cnum):
    """
    Run LPCAL in AIPS.
    
    Args:
        inname (str): a FITLD parameter inname.
        inclass (str): a FITLD parameter inclass.
        in2class (str): a FITLD parameter in2class.
        cnum (int): a FITLD parameter in2vers.
    """
    lpcal = AIPSTask('lpcal')
    lpcal.inname = inname
    lpcal.inclass = inclass
    lpcal.indisk = 1
    lpcal.inseq = 1
    if(cnum >= 1):
        lpcal.in2name = inname
        lpcal.in2class = in2class
        lpcal.in2disk = 1
        lpcal.in2vers = 1
        lpcal.in2seq = 1
        if(cnum >= 2):
            lpcal.in2vers = 2
        lpcal.cmethod = 'DFT'
        lpcal.cmodel = 'COMP'
    
    blockPrint()
    lpcal.go()
    enablePrint()


def runccedt(inname, boxfile, cnum, autoccedt):
    """
    Run CCEDT in AIPS.
    
    Args:
        inname (str): a FITLD parameter inname.
        boxfile (str): a FITLD parameter boxfile.
        cnum (int): a FITLD parameter cparm(3).
        autoccedt (boolean): if it is True, then CCEDT splits CLEAN models into sub-models automatically. If it is False, then the boxfile which specifies the locations of sub-model boxes should be provided.
    """
    ccedt = AIPSTask('ccedt')
    ccedt.inname = inname
    ccedt.inclass = 'CMAP'
    ccedt.indisk = 1
    ccedt.inseq = 1
    if (cnum == 1):
        return
    
    if autoccedt:
        ccedt.cparm[3] = cnum
        ccedt.cparm[4] = 1
    else:
        ccedt.nboxes = cnum
        dumboxfile = boxfile.split('/')[-1]
        if(len(dumboxfile) >= 46):
            newboxfile = "/".join(boxfile.split('/')[:-1]) + '/' + dumboxfile[0:45]
            os.system('cp {:} {:}'.format(boxfile, newboxfile))
            ccedt.boxfile = newboxfile
        else:
            ccedt.boxfile = boxfile
        ccedt.nccbox = -cnum
    
    blockPrint()
    ccedt.go()
    enablePrint()
    
    if not autoccedt:
        if(len(dumboxfile) >= 46):
            os.system('rm {:}'.format(newboxfile))
        

def runprtab(data, outprint):
    """
    Run PRTAB in AIPS.
    
    Args:
        data (ParselTongue UVData): an input ParselTongue UVData.
        outprint (str): the output filename of the D-terms in the antenna table.
    """
    prtab = data.table('AN', 1)

    anname = np.empty([0])
    DRArr = np.empty([0])
    DLArr = np.empty([0])
    
    ifnum = data.header.naxis[3]
    
    for row in prtab:
        anname = np.append(anname, np.repeat(row.anname.replace(' ', ''), ifnum*2))
        DRArr = np.append(DRArr, row.polcala)
        DLArr = np.append(DLArr, row.polcalb)
    
    
    df = pd.DataFrame(DRArr.transpose())
    df['antennas'] = np.array(anname)
    df['DRArr'] = np.array(DRArr)
    df['DLArr'] = np.array(DLArr)
    del df[0]
    
    df.to_csv(outprint, sep = "\t")


def runtbout(inname, inclass, outtext):
    """
    Run TBOUT in AIPS.
    
    Args:
        inname (str): a TBOUT parameter inname.
        inclass (str): a TBOUT parameter inclass.
        outtext (str): the filename of the antenna table output.
    """
    tbout = AIPSTask('tbout')
    tbout.inname = inname
    tbout.inclass = inclass
    tbout.inseq = 1
    tbout.indisk = 1
    tbout.inext = 'AN'
    tbout.invers = 1
    dumouttext = outtext.split('/')[-1]
    if(len(dumouttext) >= 46):
        newouttext = "/".join(outtext.split('/')[:-1]) + '/' + dumouttext[0:45]
        tbout.outtext = newouttext
    else:
        tbout.outtext = outtext
    tbout.docrt = -3
    blockPrint()
    tbout.go()
    enablePrint()
    
    if(len(dumouttext) >= 46):
        os.system('mv {:} {:}'.format(newouttext, outtext))
 

def runcalib(inname, inclass, in2name, in2class, outclass, solint, soltype, solmode, weightit, indisk = 1, inseq = 1, in2disk = 1, in2seq = 1, snver = 1, calsour = None):
    """
    Run CALIB in AIPS.
    
    Args:
        inname (str): a CALIB parameter inname.
        inclass (str): a CALIB parameter inclass.
        outclass (str): a CALIB parameter outclass.
        solint (float): a CALIB parameter solint.
        soltype (str): a CALIB parameter soltype.
        solmode (str): a CALIB parameter solmode.
        weightit (int): a CALIB parameter weightit.
    """
    calib = AIPSTask('calib')
    calib.inname = inname
    calib.inclass = inclass
    calib.indisk = indisk
    calib.inseq = inseq
    if(calib.in2name == 'None'):
        calib.in2name = inname
    else:
        calib.in2name = in2name
    calib.in2class = in2class
    calib.in2disk = in2disk
    calib.in2seq = in2seq
    calib.invers = 1
    calib.outclass = outclass
    calib.cmethod = 'DFT'
    calib.cmodel = 'COMP'
    calib.solint = solint
    calib.refant = 0
    if(solmode == 'a&p'):
        calib.aparm[1] = 4
    else:
        calib.aparm[1] = 3
    calib.aparm[7] = 0.1
    calib.weightit = weightit
    calib.soltype = soltype
    calib.solmode = solmode
    calib.snver = snver
    if(calsour != None):
        calib.calsour[1] = calsour
    
    blockPrint()
    calib.go()
    enablePrint()
    
                        
def runsplit(inname, inclass, inseq = 1, indisk = 1, docal = -1, gainuse = 1, sources = None, bif = 0, eif = 0, bchan = 0, echan = 0, outclass = 'SPLIT', outseq = 1, dopol = 3, aparm = 0):
    """
    Run SPLIT in AIPS to apply the D-terms.
    
    Args:
        inname (str): a SPLIT parameter inname.
        inclass (str): a SPLIT parameter inclass.
    """
    split = AIPSTask('split')
    split.inname = inname
    split.inclass = inclass
    split.inseq = inseq
    split.indisk = indisk
    split.docal = docal
    split.gainuse = gainuse
    if(sources != None):
        split.sources[1] = sources
    split.bif = bif
    split.eif = eif
    split.bchan = bchan
    split.echan = echan
    split.outclass = outclass
    split.outseq = outseq
    split.dopol = dopol
    split.aparm[1] = aparm
    blockPrint()
    split.go()
    enablePrint()



def runsplat(inname, inclass, inseq = 1, indisk = 1, outclass = 'SPLAT', docal = -1, gainuse = 1):
    """
    Run SPLIT in AIPS to apply the D-terms.
    
    Args:
        inname (str): a splat parameter inname.
        inclass (str): a splat parameter inclass.
    """
    splat = AIPSTask('splat')
    splat.inname = inname
    splat.inclass = inclass
    splat.inseq = inseq
    splat.indisk = indisk
    splat.outclass = outclass
    splat.docal = docal
    splat.gainuse = gainuse
    blockPrint()
    splat.go()
    enablePrint()



def runindxr(inname, inclass, inseq, indisk, solint):
    indxr = AIPSTask('indxr')
    indxr.inname = inname
    indxr.inclass = inclass
    indxr.inseq = inseq
    indxr.indisk = indisk
    indxr.cparm[3] = solint
    blockPrint()
    indxr.go()
    enablePrint()


def runmulti(inname, inclass, inseq, indisk):
    """
    Run MULTI in AIPS for the EVPA calibration.
    
    Args:
        inname (str): a MULTI parameter inname.
        inclass (str): a MULTI parameter inclass.
    """
    multi = AIPSTask('multi')
    multi.inname = inname
    multi.inclass = inclass
    multi.indisk = indisk
    multi.inseq = inseq
    multi.outname = inname
    blockPrint()
    multi.go()
    enablePrint()


def rundbcon(source1, inclass, inseq, indisk, source2, in2class, in2seq, in2disk, outname, outclass = 'DBCON', outseq = 1, fqcenter = -1):
        
    dbcon = AIPSTask('dbcon')
    dbcon.inname = source1
    dbcon.inclass = inclass
    dbcon.inseq = inseq
    dbcon.indisk = indisk
    
    dbcon.in2name = source2
    dbcon.in2class = in2class
    dbcon.in2seq = in2seq
    dbcon.in2disk = in2disk
    
    dbcon.outname = outname
    dbcon.outclass = outclass
    dbcon.outseq = outseq
    dbcon.fqcenter = fqcenter
    
    blockPrint()
    dbcon.go()
    enablePrint()


def runclcal(inname, inclass, indisk, inseq, snver, interpol = 'SELF', gainver = 0, gainuse = 0):
    """
    Run CLCOR in AIPS for the EVPA calibration.
    
    Args:
        inname (str): a CLCOR parameter inname.
        clcorprm (list): the list of floats of the amounts of the RCP and LCP phase offset at the reference antenna.
    """
    clcal = AIPSTask('clcal')
    clcal.inname = inname
    clcal.inclass = inclass
    clcal.indisk = indisk
    clcal.inseq = inseq
    clcal.interpol = interpol
    clcal.snver = snver
    clcal.gainver = gainver
    clcal.gainuse = gainuse
    blockPrint()
    clcal.go()
    enablePrint()
    
    

def runclcor(inname, inclass, indisk, inseq, clcorprm):
    """
    Run CLCOR in AIPS for the EVPA calibration.
    
    Args:
        inname (str): a CLCOR parameter inname.
        clcorprm (list): the list of floats of the amounts of the RCP and LCP phase offset at the reference antenna.
    """
    clcor = AIPSTask('clcor')
    clcor.inname = inname
    clcor.inclass = inclass
    clcor.indisk = indisk
    clcor.inseq = inseq
    clcor.stokes = 'L'
    clcor.opcode = 'PHAS'
    for i in range(len(clcorprm)):
        clcor.clcorprm[i+1] = float(clcorprm[i])
    blockPrint()
    clcor.go()
    enablePrint()

    
    
def runsplitpang(inname):
    """
    Run SPLIT in AIPS for the EVPA calibration.
    
    Args:
        inname (str): a SPLIT parameter inname.
    """             
    split = AIPSTask('split')
    split.inname = inname
    split.inclass = 'MULTI'
    split.indisk = 1
    split.inseq = 1
    split.dopol = -1
    split.outclass = 'PANG'
    split.docalib = 1
    split.gainuse = 2
    blockPrint()
    split.go()
    enablePrint()       


def runuvsub(inname, inclass, in2class, invers, outseq):
    """
    Run UVSUB in AIPS.
    
    Args:
        inname (str): a UVSUB parameter inname.
        inclass (str): a UVSUB parameter inclass.
        in2class (str): a UVSUB parameter in2class.
        invers (int): a UVSUB parameter invers.
        outseq (int): a UVSUB parameter outseq.
    """      
    uvsub = AIPSTask('uvsub')
    uvsub.inname = inname
    uvsub.inclass = inclass
    uvsub.indisk = 1
    uvsub.inseq = 1
    uvsub.in2name = inname
    uvsub.in2class = in2class
    uvsub.in2disk = 1
    uvsub.in2seq = 1
    uvsub.outseq = outseq
    uvsub.cmethod = 'DFT'
    uvsub.cmodel = 'COMP'
    uvsub.opcode = 'MODL'
    uvsub.invers = invers
    
    blockPrint()
    uvsub.go()
    enablePrint()



def runuvcop(inname, inclass, inseq = 1, indisk = 1, outname = None, outclass = None, fqcenter = -1):
    """
    Run UVSUB in AIPS.
    
    Args:
        inname (str): a UVSUB parameter inname.
        inclass (str): a UVSUB parameter inclass.
        in2class (str): a UVSUB parameter in2class.
        invers (int): a UVSUB parameter invers.
        outseq (int): a UVSUB parameter outseq.
    """      
    uvcop = AIPSTask('uvcop')
    uvcop.inname = inname
    uvcop.inclass = inclass
    uvcop.indisk = indisk
    uvcop.inseq = inseq
    if outname != None:
        uvcop.outname = outname
    if outclass != None:
        uvcop.outclass = outclass
    uvcop.fqcenter = fqcenter
    
    blockPrint()
    uvcop.go()
    enablePrint()




def runfittp(inname, inclass, dataout):
    """
    Run FITTP in AIPS.
    
    Args:
        inname (str): a FITTP parameter inname.
        inclass (str): a FITTP parameter inclass.
        dataout (str): the FITTP output filename.
    """
    
    if path.exists(dataout):
        os.system('rm {:}'.format(dataout))
        
    fittp = AIPSTask('fittp')
    fittp.inname = inname
    fittp.inclass = inclass
    fittp.indisk = 1
    fittp.inseq = 1
    dumdataout = dataout.split('/')[-1]
    if(len(dumdataout) >= 46):
        newdataout = "/".join(dataout.split('/')[:-1]) + '/' + dumdataout[0:45]
        fittp.dataout = newdataout
    else:
        fittp.dataout = dataout
    blockPrint()
    
    if path.exists(fittp.dataout):
        os.system('rm ' + fittp.dataout)
        
    fittp.go()
    if(len(dumdataout) >= 46):
        os.system('mv {:} {:}'.format(newdataout, dataout))
    enablePrint()


def runtacop(inname, inclass, outname, outclass):
    """
    Run TACOP in AIPS.
    
    Args:
        inname (str): a TACOP parameter inname.
        inclass (str): a TACOP parameter inclass.
        outname (str): a TACOP parameter outname.
        outclass (str): a TACOP parameter outclass.
    """
    tacop = AIPSTask('tacop')
    tacop.inname = inname
    tacop.inclass = inclass
    tacop.inseq = 1
    tacop.indisk = 1
    tacop.inext = 'AN'
    tacop.invers = 1
    tacop.outname = outname
    tacop.outclass = outclass
    blockPrint()
    tacop.go()
    enablePrint()
    


def freqavg(datain, dataout, logger = None): # Version 1.1

    """
    Run SPLIT in AIPS to apply the D-terms.
    
    Args:
        inname (str): a SPLIT parameter inname.
        inclass (str): a SPLIT parameter inclass.
    """
    
    if not (logger == None):
        logger.info('Averaging data over frequency... \n Input file: {:} \n Output file: {:}'.format(datain, dataout))
        
    data = AIPSUVData('AVG', 'EDIT', 1, 1)
    if(data.exists() == True):
        data.zap()
        
    runfitld('AVG', 'EDIT', datain)
    
    split = AIPSUVData('AVG', 'SPLIT', 1, 1)
    if(split.exists() == True):
        split.zap()
        
    split = AIPSTask('split')
    split.inname = 'AVG'
    split.inclass = 'EDIT'
    split.outclass = 'SPLIT'
    split.indisk = 1
    split.aparm[1] = 3
    blockPrint()
    split.go()
    enablePrint()
    
    runfittp('AVG', 'SPLIT', dataout)
    
    data = AIPSUVData('AVG', 'EDIT', 1, 1)
    
    data.zap()
    
    split = AIPSUVData('AVG', 'SPLIT', 1, 1)
    
    split.zap()
